#include "NVSExplorer.hpp"

// C++ system level
#include <set>
#include <cmath>
#include <limits>      // mix/max integer values; used for parsing network data
#include <vector>
#include <cstring>     // memset, memcpy, strncmp
#include <cstdlib>     // std::strtof and such…
// #include <functional>

// ESP32 specific
#include "esp_log.h"
#include "nvs_flash.h"
#include "esp_err.h"
// #include "driver/uart.h"

// project specific
#include <SpiderLib/Util.hpp>
#include "SpiderLib/NumberTypes.hpp"

// qthing stuff
#include <qthing>
#include <qthing/mqtt_common.hpp>




namespace SiliconTorch {

  namespace NVSExplorer {


    const char* TAG = "NVSExplorer";


    NVSExplorer::NVSExplorer() {

      ESP_LOGI(TAG, "Starting NVSExplorer[📡]");

      nvs_flash_init();



      u16 deviceTopicLEN = requestTopic("").length();

      auto wrapMessageHandler = [deviceTopicLEN](std::function<void(std::vector<std::string>&, qthing::multipart_message_t)> handler) {
        return [handler, deviceTopicLEN](qthing::multipart_message_t msg) {
          if (msg.length == msg.total_length && msg.offset == 0) {

            std::string localTopic(msg.topic.c_str() + deviceTopicLEN);
            std::vector<std::string> topicParts = SpiderLib::Util::split(localTopic, '/');

            handler(topicParts, msg);
          } else {
            ESP_LOGW(TAG, "Multipart messages are not supported atm.  topic[ %s ]", msg.topic.c_str());
          }
        };
      };


      // TODO: publish error if NS not found…?!
      auto lsKeys = [&](std::vector<std::string>& topic, qthing::multipart_message_t msg) {
        
        std::string delimiter = " ";
        char* buffer = NULL;

        if (msg.length > 0) {
          buffer = new char[msg.length + 1];
          snprintf(buffer, msg.length + 1, "%s", msg.payload);
          delimiter = std::string(buffer);
        }

        auto keys = listKeys(topic[1]);
        std::string msgOut = SpiderLib::Util::join(keys, delimiter);

        qthing::publish_message(responseTopic(topic[1]), msgOut);

        if (buffer != NULL) delete buffer;
      };

      qthing::add_binary_message_callback(requestTopic("ls/+"), wrapMessageHandler(lsKeys));



      auto lsNS = [&](std::vector<std::string>& topic, qthing::multipart_message_t msg) {
        
        std::string delimiter = " ";
        char* buffer = NULL;

        if (msg.length > 0) {
          buffer = new char[msg.length + 1];
          snprintf(buffer, msg.length + 1, "%s", msg.payload);
          delimiter = std::string(buffer);
        }

        auto nameSpaces = listNamespaces();
        std::string msgOut = SpiderLib::Util::join(nameSpaces, delimiter);

        qthing::publish_message(responseTopic("ls"), msgOut);

        if (buffer != NULL) delete buffer;
      };

      qthing::add_binary_message_callback(requestTopic("ls"), wrapMessageHandler(lsNS));



      auto rmKey = [&](std::vector<std::string>& topic, qthing::multipart_message_t msg) {
        this->rmKey(topic[1], topic[2]);

        // publish error…?
      };

      qthing::add_binary_message_callback(requestTopic("rm/+/+"), wrapMessageHandler(rmKey));



      auto getFloat = [&](std::vector<std::string>& topic, qthing::multipart_message_t msg) {
        float value = this->getFloat(topic[1], topic[2]);

        if (!std::isnan(value)) {

          char buffer[64];
          snprintf(buffer, 64, "%f", value);

          qthing::publish_message(responseTopic("get_float", topic[1], topic[2]), std::string(buffer));
        } else {
          // publish error…?

          ESP_LOGW(TAG, "Retrieving float[] from  namespace[ %s ]  key[ %s ]  FAILED", topic[1].c_str(), topic[2].c_str());
        }
      };

      qthing::add_binary_message_callback(requestTopic("get_float/+/+"), wrapMessageHandler(getFloat));



      auto getDouble = [&](std::vector<std::string>& topic, qthing::multipart_message_t msg) {
        double value = this->getDouble(topic[1], topic[2]);

        if (!std::isnan(value)) {

          char buffer[64];
          snprintf(buffer, 64, "%f", value);

          qthing::publish_message(responseTopic("get_double", topic[1], topic[2]), std::string(buffer));
        } else {
          // publish error…?

          ESP_LOGW(TAG, "Retrieving double[] from  namespace[ %s ]  key[ %s ]  FAILED", topic[1].c_str(), topic[2].c_str());
        }
      };

      qthing::add_binary_message_callback(requestTopic("get_double/+/+"), wrapMessageHandler(getDouble));



      auto setFloat = [&](std::vector<std::string>& topic, qthing::multipart_message_t msg) {

        char* buffer = new char[msg.length + 1];
        snprintf(buffer, msg.length + 1, "%s", msg.payload);

        float value = std::strtof(buffer, NULL);

        if ( this->setFloat(topic[1], topic[2], value) )
             ESP_LOGI(TAG, "Set  namespace[ %s ]  key[ %s ]  to float[ %f ]",             topic[1].c_str(), topic[2].c_str(), value);
        else ESP_LOGW(TAG, "Setting  namespace[ %s ]  key[ %s ]  to float[ %f ]  FAILED", topic[1].c_str(), topic[2].c_str(), value);

        // publish error…?
      };

      qthing::add_binary_message_callback(requestTopic("set_float/+/+"), wrapMessageHandler(setFloat));



      auto setDouble = [&](std::vector<std::string>& topic, qthing::multipart_message_t msg) {

        char* buffer = new char[msg.length + 1];
        snprintf(buffer, msg.length + 1, "%s", msg.payload);

        double value = std::strtod(buffer, NULL);

        if ( this->setDouble(topic[1], topic[2], value) )
             ESP_LOGI(TAG, "Set namespace[ %s ]  key[ %s ] to double[ %f ]",             topic[1].c_str(), topic[2].c_str(), value);
        else ESP_LOGW(TAG, "Setting namespace[ %s ]  key[ %s ] to double[ %f ]  FAILED", topic[1].c_str(), topic[2].c_str(), value);

        // publish error…?
      };

      qthing::add_binary_message_callback(requestTopic("set_double/+/+"), wrapMessageHandler(setDouble));




      auto getU8 = [&](std::vector<std::string>& topic, qthing::multipart_message_t msg) {

        nvs_handle_t nvs;
        if ( nvs_open(topic[1].c_str(), NVS_READONLY, &nvs) == ESP_OK ) {

          u8 value = 0;
          if ( nvs_get_u8(nvs, topic[2].c_str(), &value) == ESP_OK ) {

            char buffer[32];
            snprintf(buffer, 32, "%d", value);
            qthing::publish_message(responseTopic("get_u8", topic[1], topic[2]), std::string(buffer));

          } else {
            // publish error…?

            ESP_LOGW(TAG, "Retrieving u8[] from  namespace[ %s ]  key[ %s ]  FAILED", topic[1].c_str(), topic[2].c_str());
          }
          nvs_close(nvs);
        }
      };

      qthing::add_binary_message_callback(requestTopic("get_u8/+/+"), wrapMessageHandler(getU8));


      auto getU16 = [&](std::vector<std::string>& topic, qthing::multipart_message_t msg) {

        nvs_handle_t nvs;
        if ( nvs_open(topic[1].c_str(), NVS_READONLY, &nvs) == ESP_OK ) {

          u16 value = 0;
          if ( nvs_get_u16(nvs, topic[2].c_str(), &value) == ESP_OK ) {

            char buffer[32];
            snprintf(buffer, 32, "%d", value);
            qthing::publish_message(responseTopic("get_u16", topic[1], topic[2]), std::string(buffer));

          } else {
            // publish error…?

            ESP_LOGW(TAG, "Retrieving u16[] from  namespace[ %s ]  key[ %s ]  FAILED", topic[1].c_str(), topic[2].c_str());
          }
          nvs_close(nvs);
        }
      };

      qthing::add_binary_message_callback(requestTopic("get_u16/+/+"), wrapMessageHandler(getU16));


      auto getU32 = [&](std::vector<std::string>& topic, qthing::multipart_message_t msg) {

        nvs_handle_t nvs;
        if ( nvs_open(topic[1].c_str(), NVS_READONLY, &nvs) == ESP_OK ) {

          u32 value = 0;
          if ( nvs_get_u32(nvs, topic[2].c_str(), &value) == ESP_OK ) {

            char buffer[32];
            snprintf(buffer, 32, "%d", value);
            qthing::publish_message(responseTopic("get_u32", topic[1], topic[2]), std::string(buffer));

          } else {
            // publish error…?

            ESP_LOGW(TAG, "Retrieving u32[] from  namespace[ %s ]  key[ %s ]  FAILED", topic[1].c_str(), topic[2].c_str());
          }
          nvs_close(nvs);
        }
      };

      qthing::add_binary_message_callback(requestTopic("get_u32/+/+"), wrapMessageHandler(getU32));


      auto getU64 = [&](std::vector<std::string>& topic, qthing::multipart_message_t msg) {

        nvs_handle_t nvs;
        if ( nvs_open(topic[1].c_str(), NVS_READONLY, &nvs) == ESP_OK ) {

          u64 value = 0;
          if ( nvs_get_u64(nvs, topic[2].c_str(), &value) == ESP_OK ) {

            char buffer[32];
            snprintf(buffer, 32, "%lld", value);
            qthing::publish_message(responseTopic("get_u64", topic[1], topic[2]), std::string(buffer));

          } else {
            // publish error…?

            ESP_LOGW(TAG, "Retrieving u64[] from  namespace[ %s ]  key[ %s ]  FAILED", topic[1].c_str(), topic[2].c_str());
          }
          nvs_close(nvs);
        }
      };

      qthing::add_binary_message_callback(requestTopic("get_u64/+/+"), wrapMessageHandler(getU64));


      auto getI8 = [&](std::vector<std::string>& topic, qthing::multipart_message_t msg) {

        nvs_handle_t nvs;
        if ( nvs_open(topic[1].c_str(), NVS_READONLY, &nvs) == ESP_OK ) {

          i8 value = 0;
          if ( nvs_get_i8(nvs, topic[2].c_str(), &value) == ESP_OK ) {

            char buffer[32];
            snprintf(buffer, 32, "%d", value);
            qthing::publish_message(responseTopic("get_i8", topic[1], topic[2]), std::string(buffer));

          } else {
            // publish error…?

            ESP_LOGW(TAG, "Retrieving i8[] from  namespace[ %s ]  key[ %s ]  FAILED", topic[1].c_str(), topic[2].c_str());
          }
          nvs_close(nvs);
        }
      };

      qthing::add_binary_message_callback(requestTopic("get_i8/+/+"), wrapMessageHandler(getI8));


      auto getI16 = [&](std::vector<std::string>& topic, qthing::multipart_message_t msg) {

        nvs_handle_t nvs;
        if ( nvs_open(topic[1].c_str(), NVS_READONLY, &nvs) == ESP_OK ) {

          i16 value = 0;
          if ( nvs_get_i16(nvs, topic[2].c_str(), &value) == ESP_OK ) {

            char buffer[32];
            snprintf(buffer, 32, "%d", value);
            qthing::publish_message(responseTopic("get_i16", topic[1], topic[2]), std::string(buffer));

          } else {
            // publish error…?

            ESP_LOGW(TAG, "Retrieving i16[] from  namespace[ %s ]  key[ %s ]  FAILED", topic[1].c_str(), topic[2].c_str());
          }
          nvs_close(nvs);
        }
      };

      qthing::add_binary_message_callback(requestTopic("get_i16/+/+"), wrapMessageHandler(getI16));


      auto getI32 = [&](std::vector<std::string>& topic, qthing::multipart_message_t msg) {

        nvs_handle_t nvs;
        if ( nvs_open(topic[1].c_str(), NVS_READONLY, &nvs) == ESP_OK ) {

          i32 value = 0;
          if ( nvs_get_i32(nvs, topic[2].c_str(), &value) == ESP_OK ) {

            char buffer[32];
            snprintf(buffer, 32, "%d", value);
            qthing::publish_message(responseTopic("get_i32", topic[1], topic[2]), std::string(buffer));

          } else {
            // publish error…?

            ESP_LOGW(TAG, "Retrieving i32[] from  namespace[ %s ]  key[ %s ]  FAILED", topic[1].c_str(), topic[2].c_str());
          }
          nvs_close(nvs);
        }
      };

      qthing::add_binary_message_callback(requestTopic("get_i32/+/+"), wrapMessageHandler(getI32));


      auto getI64 = [&](std::vector<std::string>& topic, qthing::multipart_message_t msg) {

        nvs_handle_t nvs;
        if ( nvs_open(topic[1].c_str(), NVS_READONLY, &nvs) == ESP_OK ) {

          i64 value = 0;
          if ( nvs_get_i64(nvs, topic[2].c_str(), &value) == ESP_OK ) {

            char buffer[32];
            snprintf(buffer, 32, "%lld", value);
            qthing::publish_message(responseTopic("get_i64", topic[1], topic[2]), std::string(buffer));

          } else {
            // publish error…?

            ESP_LOGW(TAG, "Retrieving i64[] from  namespace[ %s ]  key[ %s ]  FAILED", topic[1].c_str(), topic[2].c_str());
          }
          nvs_close(nvs);
        }
      };

      qthing::add_binary_message_callback(requestTopic("get_i64/+/+"), wrapMessageHandler(getI64));



      auto setU8 = [&](std::vector<std::string>& topic, qthing::multipart_message_t msg) {

        char* buffer = new char[msg.length + 1];
        snprintf(buffer, msg.length + 1, "%s", msg.payload);

        u64 value = std::strtoull(buffer, NULL, 10);

        const u8 u8_max = std::numeric_limits<u8>::max();
        if (value > u8_max) value = u8_max;

        if ( this->setU8(topic[1], topic[2], static_cast<u8>(value)) )
             ESP_LOGI(TAG, "Set  namespace[ %s ]  key[ %s ]  to u8[ %lld ]",             topic[1].c_str(), topic[2].c_str(), value);
        else ESP_LOGW(TAG, "Setting  namespace[ %s ]  key[ %s ]  to u8[ %lld ]  FAILED", topic[1].c_str(), topic[2].c_str(), value);
        // publish error…?
      };

      qthing::add_binary_message_callback(requestTopic("set_u8/+/+"), wrapMessageHandler(setU8));


      auto setU16 = [&](std::vector<std::string>& topic, qthing::multipart_message_t msg) {

        char* buffer = new char[msg.length + 1];
        snprintf(buffer, msg.length + 1, "%s", msg.payload);

        u64 value = std::strtoull(buffer, NULL, 10);

        const u16 u16_max = std::numeric_limits<u16>::max();
        if (value > u16_max) value = u16_max;

        if ( this->setU16(topic[1], topic[2], static_cast<u16>(value)) )
             ESP_LOGI(TAG, "Set  namespace[ %s ]  key[ %s ]  to u16[ %lld ]",             topic[1].c_str(), topic[2].c_str(), value);
        else ESP_LOGW(TAG, "Setting  namespace[ %s ]  key[ %s ]  to u16[ %lld ]  FAILED", topic[1].c_str(), topic[2].c_str(), value);
        // publish error…?
      };

      qthing::add_binary_message_callback(requestTopic("set_u16/+/+"), wrapMessageHandler(setU16));


      auto setU32 = [&](std::vector<std::string>& topic, qthing::multipart_message_t msg) {

        char* buffer = new char[msg.length + 1];
        snprintf(buffer, msg.length + 1, "%s", msg.payload);

        u64 value = std::strtoull(buffer, NULL, 10);

        const u32 u32_max = std::numeric_limits<u32>::max();
        if (value > u32_max) value = u32_max;

        if ( this->setU32(topic[1], topic[2], static_cast<u32>(value)) )
             ESP_LOGI(TAG, "Set  namespace[ %s ]  key[ %s ]  to u32[ %lld ]",             topic[1].c_str(), topic[2].c_str(), value);
        else ESP_LOGW(TAG, "Setting  namespace[ %s ]  key[ %s ]  to u32[ %lld ]  FAILED", topic[1].c_str(), topic[2].c_str(), value);
        // publish error…?
      };

      qthing::add_binary_message_callback(requestTopic("set_u32/+/+"), wrapMessageHandler(setU32));


      auto setU64 = [&](std::vector<std::string>& topic, qthing::multipart_message_t msg) {

        char* buffer = new char[msg.length + 1];
        snprintf(buffer, msg.length + 1, "%s", msg.payload);

        u64 value = std::strtoull(buffer, NULL, 10);

        if ( this->setU64(topic[1], topic[2], value) )
             ESP_LOGI(TAG, "Set  namespace[ %s ]  key[ %s ]  to u64[ %lld ]",             topic[1].c_str(), topic[2].c_str(), value);
        else ESP_LOGW(TAG, "Setting  namespace[ %s ]  key[ %s ]  to u64[ %lld ]  FAILED", topic[1].c_str(), topic[2].c_str(), value);
        // publish error…?
      };

      qthing::add_binary_message_callback(requestTopic("set_u64/+/+"), wrapMessageHandler(setU64));


      auto setI8 = [&](std::vector<std::string>& topic, qthing::multipart_message_t msg) {

        char* buffer = new char[msg.length + 1];
        snprintf(buffer, msg.length + 1, "%s", msg.payload);

        i64 value = std::strtoll(buffer, NULL, 10);

        const i8 i8_max = std::numeric_limits<i8>::max();
        const i8 i8_min = std::numeric_limits<i8>::lowest();

        if (value > i8_max) value = i8_max;
        if (value < i8_min) value = i8_min;

        if ( this->setI8(topic[1], topic[2], static_cast<i8>(value)) )
             ESP_LOGI(TAG, "Set  namespace[ %s ]  key[ %s ]  to i8[ %lld ]",             topic[1].c_str(), topic[2].c_str(), value);
        else ESP_LOGW(TAG, "Setting  namespace[ %s ]  key[ %s ]  to i8[ %lld ]  FAILED", topic[1].c_str(), topic[2].c_str(), value);
        // publish error…?
      };

      qthing::add_binary_message_callback(requestTopic("set_i8/+/+"), wrapMessageHandler(setI8));


      auto setI16 = [&](std::vector<std::string>& topic, qthing::multipart_message_t msg) {

        char* buffer = new char[msg.length + 1];
        snprintf(buffer, msg.length + 1, "%s", msg.payload);

        i64 value = std::strtoll(buffer, NULL, 10);

        const i16 i16_max = std::numeric_limits<i16>::max();
        const i16 i16_min = std::numeric_limits<i16>::lowest();

        if (value > i16_max) value = i16_max;
        if (value < i16_min) value = i16_min;

        if ( this->setI16(topic[1], topic[2], static_cast<i16>(value)) )
             ESP_LOGI(TAG, "Set  namespace[ %s ]  key[ %s ]  to i16[ %lld ]",             topic[1].c_str(), topic[2].c_str(), value);
        else ESP_LOGW(TAG, "Setting  namespace[ %s ]  key[ %s ]  to i16[ %lld ]  FAILED", topic[1].c_str(), topic[2].c_str(), value);
        // publish error…?
      };

      qthing::add_binary_message_callback(requestTopic("set_i16/+/+"), wrapMessageHandler(setI16));


      auto setI32 = [&](std::vector<std::string>& topic, qthing::multipart_message_t msg) {

        char* buffer = new char[msg.length + 1];
        snprintf(buffer, msg.length + 1, "%s", msg.payload);

        i64 value = std::strtoll(buffer, NULL, 10);

        const i32 i32_max = std::numeric_limits<i32>::max();
        const i32 i32_min = std::numeric_limits<i32>::lowest();

        if (value > i32_max) value = i32_max;
        if (value < i32_min) value = i32_min;

        if ( this->setI32(topic[1], topic[2], static_cast<i32>(value)) )
             ESP_LOGI(TAG, "Set  namespace[ %s ]  key[ %s ]  to i32[ %lld ]",             topic[1].c_str(), topic[2].c_str(), value);
        else ESP_LOGW(TAG, "Setting  namespace[ %s ]  key[ %s ]  to i32[ %lld ]  FAILED", topic[1].c_str(), topic[2].c_str(), value);
        // publish error…?
      };

      qthing::add_binary_message_callback(requestTopic("set_i32/+/+"), wrapMessageHandler(setI32));


      auto setI64 = [&](std::vector<std::string>& topic, qthing::multipart_message_t msg) {

        char* buffer = new char[msg.length + 1];
        snprintf(buffer, msg.length + 1, "%s", msg.payload);

        i64 value = std::strtoll(buffer, NULL, 10);

        if ( this->setI64(topic[1], topic[2], value) )
             ESP_LOGI(TAG, "Set  namespace[ %s ]  key[ %s ]  to i64[ %lld ]",             topic[1].c_str(), topic[2].c_str(), value);
        else ESP_LOGW(TAG, "Setting  namespace[ %s ]  key[ %s ]  to i64[ %lld ]  FAILED", topic[1].c_str(), topic[2].c_str(), value);
        // publish error…?
      };

      qthing::add_binary_message_callback(requestTopic("set_i64/+/+"), wrapMessageHandler(setI64));







    }  // end NVSExplorer()-ctor



    u8 NVSExplorer::getU8(const std::string& nameSpace, const std::string& key, u8 defaultValue) {

      u8 value = defaultValue;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READONLY, &nvs) == ESP_OK ) {
           nvs_get_u8(nvs, key.c_str(), &value);
           nvs_close(nvs);
      }

      return value;
    }

    u16 NVSExplorer::getU16(const std::string& nameSpace, const std::string& key, u16 defaultValue) {

      u16 value = defaultValue;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READONLY, &nvs) == ESP_OK ) {
           nvs_get_u16(nvs, key.c_str(), &value);
           nvs_close(nvs);
      }

      return value;
    }

    u32 NVSExplorer::getU32(const std::string& nameSpace, const std::string& key, u32 defaultValue) {

      u32 value = defaultValue;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READONLY, &nvs) == ESP_OK ) {
           nvs_get_u32(nvs, key.c_str(), &value);
           nvs_close(nvs);
      }

      return value;
    }

    u64 NVSExplorer::getU64(const std::string& nameSpace, const std::string& key, u64 defaultValue) {

      u64 value = defaultValue;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READONLY, &nvs) == ESP_OK ) {
           nvs_get_u64(nvs, key.c_str(), &value);
           nvs_close(nvs);
      }

      return value;
    }

    i8 NVSExplorer::getI8(const std::string& nameSpace, const std::string& key, i8 defaultValue) {

      i8 value = defaultValue;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READONLY, &nvs) == ESP_OK ) {
           nvs_get_i8(nvs, key.c_str(), &value);
           nvs_close(nvs);
      }

      return value;
    }

    i16 NVSExplorer::getI16(const std::string& nameSpace, const std::string& key, i16 defaultValue) {

      i16 value = defaultValue;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READONLY, &nvs) == ESP_OK ) {
           nvs_get_i16(nvs, key.c_str(), &value);
           nvs_close(nvs);
      }

      return value;
    }

    i32 NVSExplorer::getI32(const std::string& nameSpace, const std::string& key, i32 defaultValue) {

      i32 value = defaultValue;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READONLY, &nvs) == ESP_OK ) {
           nvs_get_i32(nvs, key.c_str(), &value);
           nvs_close(nvs);
      }

      return value;
    }

    i64 NVSExplorer::getI64(const std::string& nameSpace, const std::string& key, i64 defaultValue) {

      i64 value = defaultValue;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READONLY, &nvs) == ESP_OK ) {
           nvs_get_i64(nvs, key.c_str(), &value);
           nvs_close(nvs);
      }

      return value;
    }



    bool NVSExplorer::setU8(const std::string& nameSpace, const std::string& key, u8 value) {
      bool success = true;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READWRITE, &nvs) == ESP_OK ) {
        success &= nvs_set_u8(nvs, key.c_str(), value) == ESP_OK;
        success &= nvs_commit(nvs) == ESP_OK;
                   nvs_close(nvs);

        return success;
      } else {
        return false;
      }
    }

    bool NVSExplorer::setU16(const std::string& nameSpace, const std::string& key, u16 value) {
      bool success = true;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READWRITE, &nvs) == ESP_OK ) {
        success &= nvs_set_u16(nvs, key.c_str(), value) == ESP_OK;
        success &= nvs_commit(nvs) == ESP_OK;
                   nvs_close(nvs);

        return success;
      } else {
        return false;
      }
    }

    bool NVSExplorer::setU32(const std::string& nameSpace, const std::string& key, u32 value) {
      bool success = true;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READWRITE, &nvs) == ESP_OK ) {
        success &= nvs_set_u32(nvs, key.c_str(), value) == ESP_OK;
        success &= nvs_commit(nvs) == ESP_OK;
                   nvs_close(nvs);

        return success;
      } else {
        return false;
      }
    }

    bool NVSExplorer::setU64(const std::string& nameSpace, const std::string& key, u64 value) {
      bool success = true;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READWRITE, &nvs) == ESP_OK ) {
        success &= nvs_set_u64(nvs, key.c_str(), value) == ESP_OK;
        success &= nvs_commit(nvs) == ESP_OK;
                   nvs_close(nvs);

        return success;
      } else {
        return false;
      }
    }

    bool NVSExplorer::setI8(const std::string& nameSpace, const std::string& key, i8 value) {
      bool success = true;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READWRITE, &nvs) == ESP_OK ) {
        success &= nvs_set_i8(nvs, key.c_str(), value) == ESP_OK;
        success &= nvs_commit(nvs) == ESP_OK;
                   nvs_close(nvs);

        return success;
      } else {
        return false;
      }
    }

    bool NVSExplorer::setI16(const std::string& nameSpace, const std::string& key, i16 value) {
      bool success = true;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READWRITE, &nvs) == ESP_OK ) {
        success &= nvs_set_i16(nvs, key.c_str(), value) == ESP_OK;
        success &= nvs_commit(nvs) == ESP_OK;
                   nvs_close(nvs);

        return success;
      } else {
        return false;
      }
    }

    bool NVSExplorer::setI32(const std::string& nameSpace, const std::string& key, i32 value) {
      bool success = true;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READWRITE, &nvs) == ESP_OK ) {
        success &= nvs_set_i32(nvs, key.c_str(), value) == ESP_OK;
        success &= nvs_commit(nvs) == ESP_OK;
                   nvs_close(nvs);

        return success;
      } else {
        return false;
      }
    }

    bool NVSExplorer::setI64(const std::string& nameSpace, const std::string& key, i64 value) {
      bool success = true;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READWRITE, &nvs) == ESP_OK ) {
        success &= nvs_set_i64(nvs, key.c_str(), value) == ESP_OK;
        success &= nvs_commit(nvs) == ESP_OK;
                   nvs_close(nvs);

        return success;
      } else {
        return false;
      }
    }



    float NVSExplorer::getFloat(const std::string& nameSpace, const std::string& key) {

      float value  = std::nanf("1337");
      u32*  valptr = reinterpret_cast<u32*>(&value);

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READONLY, &nvs) == ESP_OK ) {
           nvs_get_u32(nvs, key.c_str(), valptr);
           nvs_close(nvs);
      }

      return value;
    }

    double NVSExplorer::getDouble(const std::string& nameSpace, const std::string& key) {

      double value  = std::nan("1337");
      u64*   valptr = reinterpret_cast<u64*>(&value);

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READONLY, &nvs) == ESP_OK ) {
           nvs_get_u64(nvs, key.c_str(), valptr);
           nvs_close(nvs);
      }

      return value;
    }


    bool NVSExplorer::setFloat(const std::string& nameSpace, const std::string& key, float value) {

      bool success = true;
      u32  intval  = reinterpret_cast<u32*>(&value)[0];

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READWRITE, &nvs) == ESP_OK ) {
        success &= nvs_set_u32(nvs, key.c_str(), intval) == ESP_OK;
        success &= nvs_commit(nvs) == ESP_OK;
                   nvs_close(nvs);

        return success;
      } else {
        return false;
      }
    }

    bool NVSExplorer::setDouble(const std::string& nameSpace, const std::string& key, double value) {

      bool success = true;
      u64  intval  = reinterpret_cast<u64*>(&value)[0];

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READWRITE, &nvs) == ESP_OK ) {
        success &= nvs_set_u64(nvs, key.c_str(), intval) == ESP_OK;
        success &= nvs_commit(nvs) == ESP_OK;
                   nvs_close(nvs);

        return success;
      } else {
        return false;
      }
    }


    void NVSExplorer::rmKey(const std::string& nameSpace, const std::string& key) {
      nvs_handle_t _nvs;
      ESP_LOGW(TAG, "rmKey() -> %s / %s", nameSpace.c_str(), key.c_str());
      ESP_LOGW(TAG, "open() -> %s", esp_err_to_name( nvs_open(nameSpace.c_str(), NVS_READWRITE, &_nvs) ));
      ESP_LOGW(TAG, "erase() -> %s", esp_err_to_name( nvs_erase_key(_nvs, key.c_str()) ));
      ESP_LOGW(TAG, "commit() -> %s", esp_err_to_name( nvs_commit(_nvs) ));
      ESP_LOGW(TAG, "erase²() -> %s", esp_err_to_name( nvs_erase_key(_nvs, key.c_str()) ));
    }

    std::set<std::string> NVSExplorer::listKeys(const std::string& nameSpace) {

      std::set<std::string> keys;

      nvs_iterator_t it = nvs_entry_find("nvs", nameSpace.c_str(), NVS_TYPE_ANY);

      while (it != NULL) {
        nvs_entry_info_t info;
        nvs_entry_info(it, &info);
        it = nvs_entry_next(it);

        ESP_LOGW(TAG, "namespace[ %s ]   key[ %s ]   type[ %d ]", info.namespace_name, info.key, info.type);  // TODO: remove after dev!

        keys.insert(info.key);
      }

      return keys;
    }

    std::set<std::string> NVSExplorer::listNamespaces() {

      std::set<std::string> namespaces;

      nvs_iterator_t it = nvs_entry_find("nvs", NULL, NVS_TYPE_ANY);

      while (it != NULL) {
        nvs_entry_info_t info;
        nvs_entry_info(it, &info);
        it = nvs_entry_next(it);

        ESP_LOGW(TAG, "namespace[ %s ]   key[ %s ]   type[ %d ]", info.namespace_name, info.key, info.type);  // TODO: remove after dev!

        namespaces.insert(info.namespace_name);
      }

      return namespaces;
    }


    i64 getSignedInt(const std::string& nameSpace, const std::string& key, i64 defaultValue) {

      // Lets first test & evaluate the other function before duplication………

      return -1337LL;
    }


    u64 getUnsignedInt(const std::string& nameSpace, const std::string& key, u64 defaultValue) {

      u64 value = defaultValue;
      nvs_type_t entryType = NVS_TYPE_ANY;

      // search for the key to get its type
      nvs_iterator_t it = nvs_entry_find("nvs", nameSpace.c_str(), NVS_TYPE_ANY);

      while (it != NULL) {
        nvs_entry_info_t entry;
        nvs_entry_info(it, &entry);

        if (key.compare(entry.key) == 0) {
          entryType = entry.type;
          nvs_release_iterator(it);
          break;
        }

        it = nvs_entry_next(it);
      }

      if (entryType == NVS_TYPE_ANY)  // key not found
        return defaultValue;

      nvs_handle_t nvs;
      if ( nvs_open(nameSpace.c_str(), NVS_READONLY, &nvs) != ESP_OK )  // namespace not found
        return defaultValue;

      u64* valptr = &value;
      switch (entryType) {
        case NVS_TYPE_U8:  { nvs_get_u8(nvs,  key.c_str(), ( u8*)valptr); break; }
        case NVS_TYPE_I8:  { nvs_get_i8(nvs,  key.c_str(), (  i8*)valptr); break; }
        case NVS_TYPE_U16: { nvs_get_u16(nvs, key.c_str(), (u16*)valptr); break; }
        case NVS_TYPE_I16: { nvs_get_i16(nvs, key.c_str(), ( i16*)valptr); break; }
        case NVS_TYPE_U32: { nvs_get_u32(nvs, key.c_str(), (u32*)valptr); break; }
        case NVS_TYPE_I32: { nvs_get_i32(nvs, key.c_str(), ( i32*)valptr); break; }
        case NVS_TYPE_U64: { nvs_get_u64(nvs, key.c_str(), (u64*)valptr); break; }
        case NVS_TYPE_I64: { nvs_get_i64(nvs, key.c_str(), ( i64*)valptr); break; }
        default: {}
      }

      nvs_close(nvs);
      return value;
    }


    bool setUnsignedInt(const std::string& nameSpace, const std::string& key, u64 value, nvs_type_t entryType) {

      bool result = false;

      if (entryType == NVS_TYPE_ANY) {

        // check if key exists and grab its type
        nvs_iterator_t it = nvs_entry_find("nvs", nameSpace.c_str(), NVS_TYPE_ANY);

        while (it != NULL) {
          nvs_entry_info_t entry;
          nvs_entry_info(it, &entry);

          if (key.compare(entry.key) == 0) {
            entryType = entry.type;
            nvs_release_iterator(it);
            break;
          }

          it = nvs_entry_next(it);
        }

        // TODO: further implementation……?

      } else {

      }

      return result;
    }


    std::string NVSExplorer::deviceTopic(const std::string& suffix) {
      return std::string(DEVICE_NAMESPACE + "NVS/") + suffix;
    }


    std::string NVSExplorer::requestTopic(const std::string& cmd) {
      return deviceTopic( std::string("rq/") + cmd );
    }

    std::string NVSExplorer::requestTopic(const std::string& cmd, const std::string& nameSpace) {
      return deviceTopic( std::string("rq/") + cmd + std::string("/") + nameSpace );
    }

    std::string NVSExplorer::requestTopic(const std::string& cmd, const std::string& nameSpace, const std::string& key) {
      return deviceTopic( std::string("rq/") + cmd + std::string("/") + nameSpace + std::string("/") + key );
    }


    std::string NVSExplorer::responseTopic(const std::string& cmd) {
      return deviceTopic( std::string("rs/") + cmd );
    }

    std::string NVSExplorer::responseTopic(const std::string& cmd, const std::string& nameSpace) {
      return deviceTopic( std::string("rs/") + cmd + std::string("/") + nameSpace );
    }

    std::string NVSExplorer::responseTopic(const std::string& cmd, const std::string& nameSpace, const std::string& key) {
      return deviceTopic( std::string("rs/") + cmd + std::string("/") + nameSpace + std::string("/") + key );
    }

  }
}
